from pathlib import Path
from .midi_and_wav_creation import notes_to_midi, midi_to_wav
from .tabs_parser import read_tabs_file, tune_midi_notes


# tabs_path: a path to the .txt file containing the tabs input. The name of the file is the song name
# audio_save_path: a path to the directory the .midi and .wav files should be saved to
# soundfont_path: a path to the soundfont file
# save_midi (optional): if True, the .midi file will be saved, otherwise it is deleted
# play_midi (optional): if True, the .midi file will be played via the terminal on creation, otherwise nothing is played
# tempo (optional): measured in bpm, and changing the value from the default 120 will change the speed of the song
# riff_script (optional): should only be set to True if the call of run is done within RiffScript, an esoteric language that uses tabs2music as an add-on
def run(tabs_path, audio_save_path, soundfont_path, save_midi=False, play_midi=False, tempo=120, riff_script=False):
    # This returns the parsed output from the tabs file, and the song name TODO: fix this explanation
    song_name, guitar_tunings, guitar_strings = read_tabs_file(tabs_path)
    midi_notes = tune_midi_notes(guitar_tunings, guitar_strings)

    # The Path object ensures the file paths are handled as needed for any OS
    audio_save_path = Path(audio_save_path)
    audio_file_path = audio_save_path / song_name
    midi_save_path = str(audio_file_path.with_suffix('.midi'))
    wav_save_path = str(audio_file_path.with_suffix('.wav'))

    # The parsed tabs are converted to a MIDI file, and then that is converted to a wav file
    notes_to_midi(midi_notes, midi_save_path, tempo)
    midi_to_wav(soundfont_path, midi_save_path, wav_save_path, play_midi) 

    # The intermediate MIDI file is deleted by default, however the user can specify it should be saved
    if not save_midi:
        midi_path = Path(midi_save_path)
        midi_path.unlink()

    # The run function will return the parsed guitar_string notes to RiffScript as it needs the tab numbers, not the midi numbers
    if riff_script:
        return guitar_strings


import re
from pathlib import Path
from .tuning_to_midi_note import tuning_to_midi_note_mapping


# Reads the tabs file and for each guitar string, it extracts the tuning and their notes
def read_tabs_file(tabs_path):
    # The name of the tabs file is used as the song name
    song_name = extract_file_name(tabs_path)
    
    guitar_tunings = []
    guitar_strings = []

    with open(tabs_path) as tabs:
        for string in tabs:
            tuning, notes = extract_string_notes(string)
            guitar_tunings.append(tuning)
            guitar_strings.append(notes)
    
    return song_name, guitar_tunings, guitar_strings


# TODO: explain
def tune_midi_notes(guitar_tunings, guitar_strings):
    midi_notes = []

    for tuning, notes in zip(guitar_tunings, guitar_strings):
        # The fret numbers in tabs don't mean anything to MIDI, so we have to convert them
        # The midi_number is the MIDI value of the open string, and we add to it the fret number
        midi_number = tuning_to_midi_note_mapping(tuning)
        for i, note in enumerate(notes):
            if type(note) == str:
                continue
            else:
                notes[i] = note + midi_number
        midi_notes.append(notes)

    return midi_notes


# Takes a string representing a single guitar string's notes, and returns the notes and the string's tuning
def extract_string_notes(input_string):
    # This regex expression finds the tuning of the string
    tuning_pattern = r'^[a-zA-Z#b]+\d+'
    tuning = re.match(tuning_pattern, input_string).group(0)

    # This regex expressions validates the input format, so beginning with "$string tuning$ |" and ending with "|" with the notes in between
    format_pattern = r'^[a-zA-Z#b]+\d+ \|(.*)\|$'
    format_match = re.match(format_pattern, input_string)
    
    if not format_match:
        raise ValueError("Input string does not match the required format. It must begin with '$string letter$ |' and end with '|' with the notes in between.")

    # Extracts the string between the "$string tuning$ |" and "|"
    middle_part = format_match.group(1)


    notes = []
    i = 1  # Start from the second character (index 1), skips the first "-"
    while i < len(middle_part):
        # Check if it's a digit and capture the full number
        if middle_part[i].isdigit():
            num_str = middle_part[i]
            # Look ahead to capture the full number if it spans multiple characters
            while i + 1 < len(middle_part) and middle_part[i + 1].isdigit():
                i += 1
                num_str += middle_part[i]
            notes.append(int(num_str))  # Append the full number as an integer
        elif middle_part[i] in ["~", "-"]:
            notes.append(middle_part[i])  # Append "~", or "-"
        
        # Move to the next second character in the pair (skip the current '-')
        i += 2

    
    return tuning, notes


# This function returns the name of a file (without the file extension), which is used as the song name
def extract_file_name(tabs_path):
    path = Path(tabs_path)
    file_name = path.stem
    return file_name


# MIDI notes use a value between 0-127
def tuning_to_midi_note_mapping(tuning):
    note = tuning[0:-1]
    octave = int(tuning[-1])

    note_number = note_to_midi_number_mapping(note)

    # 12 is C0 for my soundfont (octave is 0, C == 0), so when its C2 ==> ((2*12) + 12) + 0 = 24 + 12 = 36
    # Note: different soundfonts may have a different number for C0 (I provided my soundfont, which can be found in the readme)
    octave_number = (octave * 12) + 12

    # This addition is similar to taking the frequency of the open string, and increasing it to the fret/note being played
    strings_midi_tuning = octave_number + note_number

    return strings_midi_tuning


# Returns a number from 0-11, where 0 == C, 1 == C#/Db, ..., 11 == B
def note_to_midi_number_mapping(note):
    if note == "C":
        return 0
    elif note == "C#" or note == "Db":
        return 1
    elif note == "D":
        return 2
    elif note == "D#" or note == "Eb":
        return 3
    elif note == "E":
        return 4
    elif note == "F":
        return 5
    elif note == "F#" or note == "Gb":
        return 6
    elif note == "G":
        return 7
    elif note == "G#" or note == "Ab":
        return 8
    elif note == "A":
        return 9
    elif note == "A#" or note == "Bb":
        return 10
    elif note == "B":
        return 11


from midiutil import MIDIFile
from midi2audio import FluidSynth


# Takes the notes defined in the tabs file and writes them to a midi file at the specified path
def notes_to_midi(guitar_strings, midi_save_path, tempo):
    # Initialize the MIDI file variables;
    track    = 0    # don't change
    channel  = 0    # don't change
    time     = 0    # In beats (don't change)
    duration = 1    # In beats (don't change)
    tempo    = tempo   # In BPM (user input)
    volume   = 127  # 0-127, as per the MIDI standard (don't change, max volume)

    MyMIDI = MIDIFile(1)  # One track, defaults to format 1 (tempo track is created automatically)
    MyMIDI.addTempo(track, time, tempo)

    # For every string, add their notes to the midi file
    for string in guitar_strings:
        # These three values are used to handle notes that can be played for longer than one beat
        duration_extension = 0
        start_time = None
        previous_pitch = None
        
        # Go through each beat for this string and add the notes as needed
        for i, note in enumerate(string):
            if note == "-": # Rest / doesn't add a note for this beat
                continue
            elif note == "~" and previous_pitch != None: # Extends the previous note played
                duration_extension += duration
            elif note == "|": # TODO: add this functionality
                pass
            else: # If the pitch is a number, then we play it
                # If a note has already been encounter before, we play it with the specified duration
                if previous_pitch != None:
                    MyMIDI.addNote(track, channel, previous_pitch, start_time, duration_extension, volume)
                
                # The below initializes the first note encountered, and resets the note for all future notes
                duration_extension = 1
                start_time = i
                previous_pitch = note
        
        # WHen all the notes have been read, check if there is a final pitch to add to the file
        if previous_pitch != None:
            MyMIDI.addNote(track, channel, previous_pitch, start_time, duration_extension, volume)


    with open(midi_save_path, "wb") as output_file:
        MyMIDI.writeFile(output_file)


# Converts the midi file into a wav file
# If the user specifies, they can hear the audio when they run the code by making play=True
def midi_to_wav(soundfont_path, midi_save_path, wav_save_path, play_midi):
    fs = FluidSynth(soundfont_path)
    fs.midi_to_audio(midi_save_path, wav_save_path)
    if play_midi:
        fs.play_midi(midi_save_path)